using System;
using System.Reflection.Emit;

namespace LinearAlgebra {

    public class Angle {
        public const float Rad2Deg = 360.0f / ((float)Math.PI * 2); //0.0174532924F;
        public const float Deg2Rad = (float)Math.PI * 2 / 360.0f; //57.29578F;

        private Angle(float degrees) {
            this.value = degrees;
        }
        private readonly float value = 0;

        public static readonly Angle zero = new(0);

        public static Angle Degrees(float degrees) {
            // Reduce it to (-180..180]
            if (float.IsFinite(degrees)) {
                while (degrees < -180)
                    degrees += 360;
                while (degrees >= 180)
                    degrees -= 360;
            }
            return new Angle(degrees);
        }

        public static Angle Radians(float radians) {
            // Reduce it to (-pi..pi]
            if (float.IsFinite(radians)) {
                while (radians <= -Math.PI)
                    radians += 2 * (float)Math.PI;
                while (radians > Math.PI)
                    radians -= 2 * (float)Math.PI;
            }

            return new Angle(radians * Rad2Deg);

        }

        public static Angle Revolutions(float revolutions) {
            // reduce it to (-0.5 .. 0.5]
            if (float.IsFinite(revolutions)) {
                // Get the integer part
                int integerPart = (int)revolutions;

                // Get the decimal part
                revolutions -= integerPart;
                if (revolutions < -0.5)
                    revolutions += 1;
                if (revolutions >= 0.5)
                    revolutions -= 1;
            }
            return new Angle(revolutions * 360);
        }

        public float inDegrees {
            get { return this.value; }
        }

        public float inRadians {
            get { return this.value * Deg2Rad; }
        }

        public float inRevolutions {
            get { return this.value / 360.0f; }
        }

        /// <summary>
        /// Get the sign of the angle
        /// </summary>
        /// <param name="a">The angle</param>
        /// <returns>-1 when the angle is negative, 1 when it is positive and 0 in all other cases</returns>
        public static int Sign(Angle a) {
            if (a.value < 0)
                return -1;
            if (a.value > 0)
                return 1;
            return 0;
        }

        /// <summary>
        /// Returns the magnitude of the angle
        /// </summary>
        /// <param name="a">The angle</param>
        /// <returns>The positive magnitude of the angle</returns>
        /// Negative values are negated to get a positive result
        public static Angle Abs(Angle a) {
            if (Sign(a) < 0)
                return -a;
            else
                return a;
        }

        /// <summary>
        /// Negate the angle
        /// </summary>
        /// <param name="a">The angle</param>
        /// <returns>The negated angle</returns>
        /// The negation of -180 is still -180 because the range is (-180..180]
        public static Angle operator -(Angle a) {
            Angle r = new(-a.value);
            return r;
        }

        /// <summary>
        /// Subtract two angles
        /// </summary>
        /// <param name="a1">Angle 1</param>
        /// <param name="a2">Angle 2</param>
        /// <returns>The result of the subtraction</returns>
        public static Angle operator -(Angle a1, Angle a2) {
            Angle r = new(a1.value - a2.value);
            return r;
        }
        /// <summary>
        /// Add two angles
        /// </summary>
        /// <param name="a1">Angle 1</param>
        /// <param name="a2">Angle 2</param>
        /// <returns>The result of the addition</returns>
        public static Angle operator +(Angle a1, Angle a2) {
            Angle r = new(a1.value + a2.value);
            return r;
        }

        /// <summary>
        /// Clamp the angle between the given min and max values
        /// </summary>
        /// <param name="angle">The angle to clamp</param>
        /// <param name="min">The minimum angle</param>
        /// <param name="max">The maximum angle</param>
        /// <returns>The clamped angle</returns>
        /// Angles are normalized
        public static float Clamp(Angle angle, Angle min, Angle max) {
            return Float.Clamp(angle.inDegrees, min.inDegrees, max.inDegrees);
        }

        /// <summary>
        /// Rotate from one angle to the other with a maximum degrees
        /// </summary>
        /// <param name="fromAngle">Starting angle</param>
        /// <param name="toAngle">Target angle</param>
        /// <param name="maxAngle">Maximum angle to rotate</param>
        /// <returns>The resulting angle</returns>
        /// This function is compatible with radian and degrees angles
        public static Angle MoveTowards(Angle fromAngle, Angle toAngle, float maxDegrees) {
            maxDegrees = Math.Max(0, maxDegrees);  // filter out negative distances
            Angle d = toAngle - fromAngle;
            float dDegrees = Abs(d).inDegrees;
            d = Degrees(Float.Clamp(dDegrees, 0, maxDegrees));
            if (Sign(d) < 0)
                d = -d;
            return fromAngle + d;
        }
    }

    /// <summary>
    /// %Angle utilities
    /// </summary>
    public static class Angles {
        public const float pi = 3.1415927410125732421875F;
        // public static float Rad2Deg = 360.0f / ((float)Math.PI * 2);
        // public static float Deg2Rad = ((float)Math.PI * 2) / 360.0f;


        /// <summary>
        /// Determine the angle difference, result is a normalized angle
        /// </summary>
        /// <param name="a">First first angle</param>
        /// <param name="b">The second angle</param>
        /// <returns>the angle between the two angles</returns>
        /// Angle values should be degrees
        public static float Difference(float a, float b) {
            float r = Normalize(b - a);
            return r;
        }

        /// <summary>
        /// Normalize an angle to the range -180 < angle <= 180
        /// </summary>
        /// <param name="angle">The angle to normalize</param>
        /// <returns>The normalized angle in interval (-180..180] </returns>
        /// Angle values should be in degrees
        public static float Normalize(float angle) {
            if (float.IsInfinity(angle))
                return angle;

            while (angle <= -180) angle += 360;
            while (angle > 180) angle -= 360;
            return angle;
        }

        /// <summary>
        /// Map interval of angles between vectors [0..Pi] to interval [0..1]
        /// </summary>
        /// <param name="v1">The first vector</param>
        /// <param name="v2">The second vector</param>
        /// <returns>The resulting factor in interval [0..1]</returns>
        /// Vectors a and b must be normalized
        /// \deprecated Please use Vector2.ToFactor instead.
        [Obsolete("Please use Vector2.ToFactor instead.")]
        public static float ToFactor(Vector2 v1, Vector2 v2) {
            return (1 - Vector2.Dot(v1, v2)) / 2;
        }

        // Normalize all vector angles to the range -180 < angle < 180
        //public static Vector3 Normalize(Vector3 angles) {
        //    float x = Normalize(angles.x);
        //    float y = Normalize(angles.y);
        //    float z = Normalize(angles.z);
        //    return new Vector3(x, y, z);
        //}

        // Returns the signed angle in degrees between from and to.
        //public static float SignedAngle(Vector3 from, Vector3 to) {
        //    float angle = Vector3.Angle(from, to);
        //    Vector3 cross = Vector3.Cross(from, to);
        //    if (cross.y < 0) angle = -angle;
        //    return angle;
        //}

        // Returns the signed angle in degrees between from and to.
        //public static float SignedAngle(Vector2 from, Vector2 to) {
        //    float sign = Math.Sign(from.y * to.x - from.x * to.y);
        //    return Vector2.Angle(from, to) * sign;
        //}

        //public static Quaternion ToQuaternion(Rotation orientation) {
        //    return new Quaternion(orientation.x, orientation.y, orientation.z, orientation.w);
        //}
    }
}